@@ -2,19 +2,62 @@
 
 /** @generate-class-entries */
 
-namespace FTP {
-
-    /**
-     * @strict-properties
-     * @not-serializable
-     */
-    final class Connection
-    {
-    }
-
-}
-
 namespace {
+    /**
+     * @var int
+     * @cvalue FTPTYPE_ASCII
+     */
+    const FTP_ASCII = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue FTPTYPE_ASCII
+     */
+    const FTP_TEXT = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue FTPTYPE_IMAGE
+     */
+    const FTP_BINARY = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue FTPTYPE_IMAGE
+     */
+    const FTP_IMAGE = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue PHP_FTP_AUTORESUME
+     */
+    const FTP_AUTORESUME = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue PHP_FTP_OPT_TIMEOUT_SEC
+     */
+    const FTP_TIMEOUT_SEC = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue PHP_FTP_OPT_AUTOSEEK
+     */
+    const FTP_AUTOSEEK = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue PHP_FTP_OPT_USEPASVADDRESS
+     */
+    const FTP_USEPASVADDRESS = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue PHP_FTP_FAILED
+     */
+    const FTP_FAILED = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue PHP_FTP_FINISHED
+     */
+    const FTP_FINISHED = UNKNOWN;
+    /**
+     * @var int
+     * @cvalue PHP_FTP_MOREDATA
+     */
+    const FTP_MOREDATA = UNKNOWN;
 
     function ftp_connect(string $hostname, int $port = 21, int $timeout = 90): FTP\Connection|false {}
 
@@ -22,7 +65,7 @@ namespace {
     function ftp_ssl_connect(string $hostname, int $port = 21, int $timeout = 90): FTP\Connection|false {}
     #endif
 
-    function ftp_login(FTP\Connection $ftp, string $username, string $password): bool {}
+    function ftp_login(FTP\Connection $ftp, string $username, #[\SensitiveParameter] string $password): bool {}
     function ftp_pwd(FTP\Connection $ftp): string|false {}
     function ftp_cdup(FTP\Connection $ftp): bool {}
     function ftp_chdir(FTP\Connection $ftp, string $directory): bool {}
@@ -67,7 +110,7 @@ namespace {
     function ftp_nb_fget(FTP\Connection $ftp, $stream, string $remote_filename, int $mode = FTP_BINARY, int $offset = 0): int {}
     function ftp_pasv(FTP\Connection $ftp, bool $enable): bool {}
     function ftp_get(FTP\Connection $ftp, string $local_filename, string $remote_filename, int $mode = FTP_BINARY, int $offset = 0): bool {}
-    function ftp_nb_get(FTP\Connection $ftp, string $local_filename, string $remote_filename, int $mode = FTP_BINARY, int $offset = 0): int|false {}
+    function ftp_nb_get(FTP\Connection $ftp, string $local_filename, string $remote_filename, int $mode = FTP_BINARY, int $offset = 0): int {}
     function ftp_nb_continue(FTP\Connection $ftp): int {}
 
     /** @param resource $stream */
@@ -91,5 +134,14 @@ namespace {
     /** @param int|bool $value */
     function ftp_set_option(FTP\Connection $ftp, int $option, $value): bool {}
     function ftp_get_option(FTP\Connection $ftp, int $option): int|bool {}
+}
 
+namespace FTP {
+    /**
+     * @strict-properties
+     * @not-serializable
+     */
+    final class Connection
+    {
+    }
 }
